/* Auteur - La Massiot - 2005 */

/*______________________________________*/
/*--------------------------------------*/
/*               cache.c                */
/*______________________________________*/
/*--------------------------------------*/


#include "types.h"
#include "cache.h"
#include <string.h>
#include <strings.h>

/* Several local functions */
error_t read_sblock_aux(sblocknum_t, sblock_t);
error_t write_sblock_aux(sblocknum_t, sblock_t);
error_t self_test_cache(int, int);

/* Several local variables */
                     int   size_of_cache;
                     int   size_of_hashtable;
struct s_list_elem_cache   cache;
struct s_list_elem_cache * hashtable;



/*-------------------------------------------------------------------------*/
/*                               init_cache                                */
/*-------------------------------------------------------------------------*/
error_t init_cache(int nb_total_blocks)
{ 
error_t e;
int i;
struct s_list_elem_cache initalisator;

  if((e = start_disk()) != 0)
    return e;
    
  initalisator.p_first_elem_cache = NULL;
  initalisator.p_last_elem_cache  = NULL;
  
  size_of_cache     = max(5, nb_total_blocks/FRACTION_CACHEE_DU_DISQUE);
  size_of_hashtable = max(5, size_of_cache/TAUX_DE_REDUCTION_HASHTABLE);

  hashtable = (struct s_list_elem_cache *) malloc(sizeof(struct s_list_elem_cache)*size_of_hashtable);

  if(hashtable == NULL)
  {
    if((e = stop_disk()) != 0)
      return e;
    
    return ERROR_MALLOC;
  }
    
  cache.p_first_elem_cache = NULL;
  cache.p_last_elem_cache  = NULL;
  cache.nb_of_elements_in_cache = 0;

  for(i=0 ; i<size_of_hashtable ; i++)
    hashtable[i] = initalisator;

  return 0;
}
/*-------------------------------------------------------------------------*/
/*                            shutdown_cache                               */
/*-------------------------------------------------------------------------*/
error_t shutdown_cache()
{
error_t e1 = 0;
error_t e2;

  while(cache.p_first_elem_cache != NULL) 
  {
    if((e2 = suppr_last_elem_cache()) != 0)
      e1 = e2;
  }
 
  free(hashtable);

  return e1;
}
/*-------------------------------------------------------------------------*/
/*                              flush_cache                                */
/*-------------------------------------------------------------------------*/
/* 
Cette fonction est  peu prs identique  shutdown_cache, sauf qu'elle ne 
fait que le vider.
*/
error_t flush_cache()
{
error_t e1 = 0;
error_t e2;

  while(cache.p_first_elem_cache != NULL) 
  {
    if((e2 = suppr_last_elem_cache()) != 0)
      e1 = e2;
  }
 
  return e1;
}
/*-------------------------------------------------------------------------*/
/*                         enlever_element_de_la_liste                        */
/*-------------------------------------------------------------------------*/
void enlever_element_de_la_liste(struct s_elem_cache * b)
{
struct s_elem_cache * a;
struct s_elem_cache * c;

  a = b->p_prev_elem_cache;
  c = b->p_next_elem_cache;

  if(a == NULL)
  {
    /* tte */
    cache.p_first_elem_cache = c;
    
    if(c == NULL) 
      cache.p_last_elem_cache = NULL;
    
    else 
      c->p_prev_elem_cache = NULL;
  } 
  
  else if (c == NULL) 
  {
    /* fin */
    cache.p_last_elem_cache = a;
    
    if(a == NULL) 
      cache.p_first_elem_cache = NULL;
    
    else 
      a->p_next_elem_cache = NULL;
  } 
  
  else 
  {
    /* milieu */

    a->p_next_elem_cache = c;
    c->p_prev_elem_cache = a;
  }

  b->p_prev_elem_cache = NULL;
  b->p_next_elem_cache = NULL;
}
/*-------------------------------------------------------------------------*/
/*                             suppr_last_elem_cache                       */
/*-------------------------------------------------------------------------*/
error_t suppr_last_elem_cache()
{
struct s_elem_cache * p_prov_elem_hash;
int location_in_hashtable;
int nro_last_elem_cache;
error_t e = 0; 
  
  p_prov_elem_hash = cache.p_last_elem_cache ;
  
  if(p_prov_elem_hash == NULL)
    return 0;
  
  nro_last_elem_cache = p_prov_elem_hash->nro_block;
  location_in_hashtable = nro_last_elem_cache % size_of_hashtable;
  
  /* On enlve l'lment de la liste du hash. */
  if((p_prov_elem_hash->p_prev_elem_hash == NULL) && (p_prov_elem_hash->p_next_elem_hash != NULL))
  {
    /* Il est en dbut de liste et il y a quelqu'un d'autre dans la liste. */
     hashtable[location_in_hashtable].p_first_elem_cache = p_prov_elem_hash->p_next_elem_hash;
     (p_prov_elem_hash->p_next_elem_hash)->p_prev_elem_hash = NULL; 
  }
  
  else if((p_prov_elem_hash->p_prev_elem_hash != NULL) && (p_prov_elem_hash->p_next_elem_hash == NULL))
  {
    /* Il est le dernier mais il y en a d'autres avant. */
    (p_prov_elem_hash->p_prev_elem_hash)->p_next_elem_hash = NULL;
    hashtable[location_in_hashtable].p_last_elem_cache = p_prov_elem_hash->p_prev_elem_hash;
  }

  else if((p_prov_elem_hash->p_prev_elem_hash == NULL) && (p_prov_elem_hash->p_next_elem_hash == NULL))
  {
    /* Il est seul dans la liste. */
     hashtable[location_in_hashtable].p_first_elem_cache = NULL;
     hashtable[location_in_hashtable].p_last_elem_cache = NULL;
  }
  
  else
  {
    /* Il est au milieu de la liste. */
    (p_prov_elem_hash->p_prev_elem_hash)->p_next_elem_hash = p_prov_elem_hash->p_next_elem_hash;
    (p_prov_elem_hash->p_next_elem_hash)->p_prev_elem_hash = p_prov_elem_hash->p_prev_elem_hash;
  }

  cache.nb_of_elements_in_cache --;

  /* On enlve l'lment de la liste L. */
  enlever_element_de_la_liste(cache.p_last_elem_cache);
  
  if(p_prov_elem_hash->b_to_be_written_physically == 1)
    e = write_sblock_aux(p_prov_elem_hash->nro_block, (sblock_t) p_prov_elem_hash->datas);
    
  free(p_prov_elem_hash); 

  return e;
}
/*-------------------------------------------------------------------------*/
/*                   ajouter_element_en_tete_de_liste                      */
/*-------------------------------------------------------------------------*/
void ajouter_element_en_tete_de_liste(struct s_elem_cache * b)
{
struct s_elem_cache * c;

  c = cache.p_first_elem_cache;
  
  if(c == NULL) 
  {
    cache.p_first_elem_cache = b;
    cache.p_last_elem_cache = b;
    b->p_next_elem_cache = NULL;
    b->p_prev_elem_cache = NULL;
  } 
  
  else 
  {
    cache.p_first_elem_cache = b;
    b->p_next_elem_cache = c;
    b->p_prev_elem_cache = NULL;
    c->p_prev_elem_cache = b;
  }
}
/*-------------------------------------------------------------------------*/
/*                            ajouter_dans_hashtable                       */
/*-------------------------------------------------------------------------*/
void ajouter_dans_hashtable(struct s_elem_cache * p_new_elem_cache)
{
int location_in_hashtable;

  location_in_hashtable = (p_new_elem_cache->nro_block) % size_of_hashtable;
  
  p_new_elem_cache->p_prev_elem_hash = NULL;
  p_new_elem_cache->p_next_elem_hash = hashtable[location_in_hashtable].p_first_elem_cache;

  if(hashtable[location_in_hashtable].p_first_elem_cache == NULL)
    hashtable[location_in_hashtable].p_last_elem_cache = p_new_elem_cache;

  else
    hashtable[location_in_hashtable].p_first_elem_cache->p_prev_elem_hash = p_new_elem_cache;

  hashtable[location_in_hashtable].p_first_elem_cache = p_new_elem_cache;
}
/*-------------------------------------------------------------------------*/
/*                                  update_cache                           */
/*-------------------------------------------------------------------------*/
error_t update_cache(int nro_block, int b_write, sblock_t block)
{
int location_in_hashtable;
struct s_elem_cache * p_elem_cache;
struct s_elem_cache * p_new_elem_cache;
int b_is_cached;
error_t e = 0;

  location_in_hashtable = nro_block % size_of_hashtable;
  p_elem_cache = hashtable[location_in_hashtable].p_first_elem_cache;
  b_is_cached = 0;

  /* On regarde si le bloc que l'on veut cacher est dj dans
     le cache ou pas. 
     Si le bloc est dj cach, on met le boolen "b_is_cached" 
      true (1).
  */
  while(p_elem_cache != NULL)
  {
    if(p_elem_cache->nro_block == nro_block)
    {
      b_is_cached = 1;
      break;
    }
    p_elem_cache = p_elem_cache->p_next_elem_hash;
  }



  
  /*--------- Le bloc est dj cach ----------*/
  if(b_is_cached == 1)
  {
    /* On le dplace en tte de "cache" */
    enlever_element_de_la_liste(p_elem_cache);
    ajouter_element_en_tete_de_liste(p_elem_cache);

    /* On crit "block" dans le cache */
    if(b_write == 1)
    {
      memcpy(p_elem_cache->datas, block, SIZE_OF_SBLOCK);
      p_elem_cache->b_to_be_written_physically = 1;
    }
        
    /* On retourne dans "block" les donnes demandes et lues
       dans le cache. */
    else
      memcpy(block, p_elem_cache->datas, SIZE_OF_SBLOCK);
  }


  
  /*------------ Le bloc n'est pas dj cach ---------------*/
  else
  {
    /* On cr un nouvel lment pour le cache destin 
       contenir les donnes du bloc de numro "nro_block" */ 
     p_new_elem_cache = (struct s_elem_cache *) malloc(sizeof(struct s_elem_cache));

     if(p_new_elem_cache == NULL)
       return ERROR_MALLOC;
     
     /* On affecte son champ "nro_block" */
     p_new_elem_cache->nro_block = nro_block;

     /*----- Lecture -----*/
     if(b_write == 0)
     {
       /* On lit le bloc (physiquement) */
       e = read_sblock_aux(nro_block, p_new_elem_cache->datas);
       
       if(e != 0)
       {
         free(p_new_elem_cache);
         return e;
       }
       
       /* On copie les donnes demandes en lecture dans "block".
          Avec le flag "b_to_be_written_physically", on 
          indique que le bloc n'a pas t modifi (dans le cache). */
       memcpy(block, p_new_elem_cache->datas, SIZE_OF_SBLOCK);
       p_new_elem_cache->b_to_be_written_physically = 0;
     }
     
     /*----- Ecriture -----
        Le bloc est crit dans le cache.
        Avec le flag "b_to_be_written_physically", on mmorise
        le fait que le bloc a t modifi (il devra donc tre
        crit physiquement lorsqu'il sera supprim du cache). */
     else
     {
       memcpy(p_new_elem_cache->datas, block, SIZE_OF_SBLOCK);
       p_new_elem_cache->b_to_be_written_physically = 1;
     }
       
     /* On rajoute l'lment dans "hashtable" */
     ajouter_dans_hashtable(p_new_elem_cache);
 
     /* On rajoute l'lment en tte de "cache" */
     ajouter_element_en_tete_de_liste(p_new_elem_cache);     

     /* Le nombre de blocs cachs augmente d'une unit */
     cache.nb_of_elements_in_cache++;

     /* Le cache est trop plein, il faut supprimer son dernier
        lment pour faire de la place. */
     if(cache.nb_of_elements_in_cache == size_of_cache)
       suppr_last_elem_cache();
  }
 
  return 0;
}
/*-------------------------------------------------------------------------*/
/*                               read_sblock_aux                           */
/*-------------------------------------------------------------------------*/
/* 
Si au niveau du cache je suis en train de lire le block de numro "num" 
alors dans la couche physique, je suis en train de lire les blocks de numros 
"2*num" et "2*num + 1". 
*/

error_t read_sblock_aux(sblocknum_t num, sblock_t block)
{
error_t e;

  if((e = read_pblock(2*num, block)) != 0)
    return (e);
  
  e = read_pblock(2*num + 1, block + 128);

  return (e);
}
/*-------------------------------------------------------------------------*/
/*                              write_sblock_aux                           */
/*-------------------------------------------------------------------------*/
error_t write_sblock_aux(sblocknum_t num , sblock_t block)
{
error_t e;

  if((e = write_pblock(2*num, block)) != 0)
    return (e);
  
  e = write_pblock(2*num + 1, block + 128); 

  return (e);
}
/*-------------------------------------------------------------------------*/
/*                               read_sblock                               */
/*-------------------------------------------------------------------------*/
error_t read_sblock(sblocknum_t num, sblock_t block)
{
  return update_cache(num, 0, block);
}
/*-------------------------------------------------------------------------*/
/*                               write_sblock                              */
/*-------------------------------------------------------------------------*/
error_t write_sblock(sblocknum_t num, sblock_t block)
{
  return update_cache(num, 1, block);
}
/*-------------------------------------------------------------------------*/
/*                                   max                                   */
/*-------------------------------------------------------------------------*/
int max(int e1, int e2)
{
  if(e1 < e2)
    return e2;

  return e1;
}
/*-------------------------------------------------------------------------*/
/*                           Self-test cache                               */
/*-------------------------------------------------------------------------*/
struct s_block_test
{
  u32 entier_test;
  char remplissage [252];
};

error_t self_test_cache(int m, int nb_test)
{
int random_int; /* Utilis pour le remplissage initial */
int random_op;  /* Pour choisir au hasard une opration entre criture/lecture/flush */
int random_int_to_be_written;  /* Dans le cas d'une criture, entier  crire au hasard */
int random_bloc_to_be_written; /* Dans le cas d'une criture, bloc  crire au hasard */
int random_bloc_to_be_read;    /* Dans le cas d'une lecture, bloc  lire au hasard */
  

int tableau [m];
struct s_block_test bloc_test;

int i;
error_t e;

  bzero(bloc_test.remplissage, sizeof(bloc_test.remplissage));

  /* INITIALISATION */
  for(i=0 ; i<m ; i++)
  {
    /* On choisit un entier au hasard */
    random_int = rand() % m;
   
    /* On remplit le tableau et le champ entier_test du bloc */
    tableau[i] = random_int;
    bloc_test.entier_test = random_int;

    /* On crit le bloc dans le cache */
    if((e = write_sblock_aux(i, (sblock_t) &bloc_test)) != 0)
      return e;
  }
  
  /* OPERATIONS */
  for(i=0 ; i<nb_test ; i++)
  {
    random_op = rand() % 3;

    if(random_op == 0) /* Opration d'criture */
    {
      /* On calcule au hasard le numro du bloc qui va tre crit et l'entier
         qui va tre crit */
      random_int_to_be_written = rand() % m;
      random_bloc_to_be_written = rand() % m;
      
      /* On crit dans le tableau */
      tableau[random_bloc_to_be_written] = random_int_to_be_written;
      
      /* On crit sur le disque */
      bloc_test.entier_test = random_int_to_be_written;

      if((e = write_sblock_aux(random_bloc_to_be_written, (sblock_t) &bloc_test)) != 0)
        return e;
    }

    if(random_op == 1) /* Opration de lecture */
    {
      /* On calcule au hasard le numro du bloc qui va tre lu */
      random_bloc_to_be_read = rand() % m;
      
      /* On lit le boc */
      if((e = read_sblock_aux(random_bloc_to_be_read, (sblock_t) &bloc_test)) != 0)
        return e;
     
      /* On compare les valeurs dans le tableau et dans le bloc de l'entier crit au pralable */
      if(tableau[random_bloc_to_be_read] != bloc_test.entier_test)
      {
        fprintf(stderr, "Problme de lecture avec le cache dans le bloc %d.\n", random_bloc_to_be_read);
        fprintf(stderr, "On s'attendait  ce que %d = %ld.\n", tableau[random_bloc_to_be_read], bloc_test.entier_test);
        return(ERROR_CACHE_TEST);
      }
    }
    
    else /* flush the cache */
    {
      if((e = flush_cache()) != 0)
      {
        fprintf(stderr, "Problme de flush avec le cache\n");
        return(ERROR_CACHE_TEST);
      }
    }
  }

  /* VERIFICATION FINALE DE LA COHERENCE */
  for(i=0 ; i<m ; i++)
  {
    /* On lit le boc */
    if((e = read_sblock_aux(i, (sblock_t) &bloc_test)) != 0)
      return e;
    
    /* On compare les valeurs dans le tableau et dans le disque */
    if(tableau[i] != bloc_test.entier_test)
    {
      fprintf(stderr, "L'entier %d du tableau et l'entier correspondant dans le bloc %d du disque ne concident pas.\n", i, i);
      fprintf(stderr, "On s'attendait  ce que %d = %ld.\n", tableau[i], bloc_test.entier_test);
      return(ERROR_CACHE_TEST);
    }
  }
  
  return 0;
}
