/* Auteur - La Massiot - 2005 */

/*_________________________________________*/
/*-----------------------------------------*/
/*         couche_applications.c           */
/*_________________________________________*/
/*-----------------------------------------*/

#include "couche_logique.h"
#include <string.h>
#include <strings.h>

void my_write_sblock(sblocknum_t, sblock_t);
void my_read_sblock(sblocknum_t, sblock_t);
void format(int, int);

int cat_fs(char * name);
int cat_into(char * name);
int mkdir_fs(char * name);
int link_fs(char * old_path, char * new_path);
int cp_fs(char * src, char * dest);
int rm_fs(char * path);
int ls_fs(char * dir);




/*-------------------------------------------------------------------------------*/
int main(int argc, char ** argv)
{
 error_t e;
 int code_exit;

  /* FORMATER DISQUE
     ./fs format 45 300 */
  if((argc == 4) && (strcmp(argv[1], "format") == 0))
  {
    init_cache(atoi(argv[2]) + atoi(argv[3]) + 1);

    format(atoi(argv[2]), atoi(argv[3]));
    
    e = shutdown_cache();

    if(e != 0)
    {
      fprintf(stderr, "main : Fatal Error shutdown_cache : %s.\n", string_of_error(e));
      exit(EXIT_FAILURE);
    }
  }
  
  /* CAT 
     ./fs cat ref_abs */
  else if((argc == 3) && (strcmp(argv[1], "cat") == 0))
  {
    if((e = init_couche_logique()) != 0)
    {
      fprintf(stderr, "main : Fatal Error init_couche_logique : %s.\n", string_of_error(e));
      exit(EXIT_FAILURE);
    }

    code_exit = cat_fs(argv[2]);
    
    shutdown_couche_logique();
    exit(code_exit);
  }
  
  /* CAT_INTO
     ./fs cat_into <ref_abs> */
  else if((argc == 3) && (strcmp(argv[1], "cat_into") == 0))
  {
    if((e = init_couche_logique()) != 0)
    {
      fprintf(stderr, "main : Fatal Error init_couche_logique : %s.\n", string_of_error(e));
      exit(EXIT_FAILURE);
    }

    code_exit = cat_into(argv[2]);
    
    shutdown_couche_logique();        
    exit(code_exit);
  }
  
  /* LINK
     ./fs link <ref_abs_1> <ref_abs_2> */
  else if((argc == 4) && (strcmp(argv[1], "link") == 0))
  {
    if((e = init_couche_logique()) != 0)
    {
      fprintf(stderr, "main : Fatal Error init_couche_logique : %s.\n", string_of_error(e));
      exit(EXIT_FAILURE);
    }

    code_exit = link_fs(argv[2], argv[3]);
    
    shutdown_couche_logique();        
    exit(code_exit);
  }

  /* MKDIR 
     ./fs mkdir <ref_abs> */
  else if((argc == 3) && (strcmp(argv[1], "mkdir") == 0))
  {
    if((e = init_couche_logique()) != 0)
    {
      fprintf(stderr, "main : Fatal Error init_couche_logique : %s.\n", string_of_error(e));
      exit(EXIT_FAILURE);
    }

    code_exit = mkdir_fs(argv[2]);
    
    shutdown_couche_logique();        
    exit(code_exit);
  }
  
  /*RM
    ./fs rm <ref_abs>*/
  else if((argc == 3) && (strcmp(argv[1], "rm") == 0))
  {
    if((e = init_couche_logique()) != 0)
    {
      fprintf(stderr, "main : Fatal Error init_couche_logique : %s.\n", string_of_error(e));
      exit(EXIT_FAILURE);
    }

    code_exit = rm_fs(argv[2]);
    
    shutdown_couche_logique();        
    exit(code_exit);
  }

  /* LS
     ./fs ls <ref_abs> */
  else if((argc == 3) && (strcmp(argv[1], "ls") == 0))
  {
    if((e = init_couche_logique()) != 0)
    {
      fprintf(stderr, "main : Fatal Error init_couche_logique : %s.\n", string_of_error(e));
      exit(EXIT_FAILURE);
    }

    code_exit = ls_fs(argv[2]);
    
    shutdown_couche_logique();        
    exit(code_exit);
  }
  
  /* CP
     ./fs cp <ref_abs_1> <ref_abs_2> */
  else if((argc == 4) && (strcmp(argv[1], "cp") == 0))
  {
    if((e = init_couche_logique()) != 0)
    {
      fprintf(stderr, "main : Fatal Error init_couche_logique : %s.\n", string_of_error(e));
      exit(EXIT_FAILURE);
    }

    code_exit = cp_fs(argv[2], argv[3]);
    
    shutdown_couche_logique();        
    exit(code_exit);
  }

  /* Self Test Cache */
  else if((argc == 4) && (strcmp(argv[1], "self_test_cache") == 0))
  {
    init_cache(atoi(argv[2]));

    e = self_test_cache(atoi(argv[2]), atoi(argv[3]));
    
    if(e != 0)
    {
      fprintf(stderr, "main : Fatal Error self_test_cache : %s.\n", string_of_error(e));
      exit(EXIT_FAILURE);
    }
    
    e = shutdown_cache();

    if(e != 0)
    {
      fprintf(stderr, "main : Fatal Error shutdown_cache : %s.\n", string_of_error(e));
      exit(EXIT_FAILURE);
    }
  }

  /* Tests read and write logiques */
  else if((argc == 4) && (strcmp(argv[1], "test_rw_logiques") == 0))
  {
    if((e = init_couche_logique()) != 0)
    {
      fprintf(stderr, "main : Fatal Error init_couche_logique : %s.\n", string_of_error(e));
      exit(EXIT_FAILURE);
    }
    
    if(test_read_write_logiques(atoi(argv[2]), atoi(argv[3])) != 0) /* 2560 */
    { 
      shutdown_couche_logique();        
      fprintf(stderr, "main : Fatal Error test_read_write_logiques.\n");
      exit(EXIT_FAILURE);
    }
  }

  else
  {
    printf("Commande inconnue ou mal forme : essayez : \n");
    printf(" format <nb_files> <nb_data_blocks>\n");
    printf(" ls </path>\n");
    printf(" cat </path>\n");
    printf(" cat_into </path>\n");
    printf(" cp </src> </dest>\n");
    printf(" rm </path> \n");
    printf(" link </old_path> </new_path>\n");
    printf(" mkdir </path> \n");
  }
  
  return 0;
}
/*------------------------------------------------*/
/*                  my_write_sblock               */
/*------------------------------------------------*/

void my_write_sblock(sblocknum_t nro_block, sblock_t block)
{
error_t e;

  e = write_sblock(nro_block, block);
  
  if(e != 0)
  {
    fprintf(stderr, "format : Fatal Error write_sblock : %s in block number %d.\n", 
            string_of_error(e), nro_block);
    exit(EXIT_FAILURE);
  }
}
/*---------------------------------------------------*/
/*                   my_read_sblock                  */
/*---------------------------------------------------*/
void my_read_sblock(sblocknum_t nro_block, sblock_t block)
{
error_t e;

  e = read_sblock(nro_block, block);
  
  if(e != 0)
  {
    fprintf(stderr, "format : Fatal Error read_sblock : %s in block number %d.\n", string_of_error(e), nro_block);
    exit(EXIT_FAILURE);
  }
}
/*---------------------------------------------*/
/*                    format                   */
/*---------------------------------------------*/
void format(int n, int d)
{
struct s_partition_de_description pd;
struct s_fs_inode super_inode;
struct s_fs_inode inode;
int i;
char array_of_size_of_sblocks_zeros [SIZE_OF_SBLOCK];
 
  bzero(&pd, sizeof(pd));
  /* Partition de description */
  pd.pd_n                    = u32_to_big_endian(n);
  pd.pd_nb_of_free_files     = u32_to_big_endian(n);
  pd.pd_nb_of_free_blocks    = u32_to_big_endian(d);
  pd.pd_nro_first_free_block = u32_to_big_endian(n+1);

  bzero(&super_inode, sizeof(super_inode));
  /* Super bloc, premier bloc de la partition des fichiers,
     rpertoire racine */
  super_inode.pf_type  = TYPE_DIRECTORY;
  super_inode.pf_protect = PERMIT_READ | PERMIT_WRITE;
  super_inode.pf_size = 0;
  super_inode.pf_state = u16_to_big_endian(STATE_OCCUPIED);
  super_inode.pf_links = u16_to_big_endian(1);
  
  /* Write the first two blocks in the cache */
  my_write_sblock(0, (sblock_t) &pd);
  my_write_sblock(1, (sblock_t) &super_inode);
 
  bzero(&inode, sizeof(inode));
  /* Partition des fichier, les (n - 1) blocs suivants */
  inode.pf_state = u16_to_big_endian(STATE_FREE);

  for(i=2 ; i<=n ; i++)
    my_write_sblock(i, (sblock_t) &inode);
    
  for(i=0 ; i<SIZE_OF_SBLOCK ; i++)
  {
    array_of_size_of_sblocks_zeros[i] = 'x';
  }
  
  /* Data Partition : d blocks filed with zeros characters : 0 or '\0' or '\x00'
     (hexa)
  */
  for(i=n+1 ; i<=n+d ; i++)
    my_write_sblock(i, (sblock_t) array_of_size_of_sblocks_zeros);
}
/*------------------------------------*/
/*               cat_fs               */
/*------------------------------------*/
/*
Cette fonction permet de copier sur la sortie standard
le contenu d'un fichier.
*/
int cat_fs(char * name)
{
fd file_descr;
int nb_bytes_read;
char buffer [1024];
error_t e;
   
  /* OUVERTURE DU FICHIER EN LECTURE SEULE ("1") */
  if((e = open_file(name, 1, 0, 0, 0, 0, &file_descr)) != 0)
  {
    fprintf(stderr, "Error in opening file %s. %s\n", name, string_of_error(e));
    return(EXIT_FAILURE);
  }

  do
  {
    /* LECTURE DU FICHIER
       On rcupre dans buffer ce qui a t lu */
    if((e = read_file(file_descr, buffer, sizeof(buffer), &nb_bytes_read)) != 0) 
    {
      fprintf(stderr, "\"read_file\" : Error in reading file %s. %s\n", name, string_of_error(e));
      close_file(file_descr);
      return(EXIT_FAILURE);
    }

    /* ECRITURE SUR LA SORTIE STANDARD DE buffer */
    if(write(STDOUT_FILENO, buffer, nb_bytes_read) == -1)
    {
      fprintf(stderr, "Error in writing (\"write\" POSIX).");
      close_file(file_descr);
      return(EXIT_FAILURE);
    }
  }
  while(nb_bytes_read != 0);
  
  return(EXIT_SUCCESS);
}
/*------------------------------*/
/*          cat_into            */ 
/*------------------------------*/
/*
Cette fonction permet de copier l'entre standard dans un fichier.
*/
int cat_into(char * name)
{
fd file_descr;
int nb_bytes_read;  
char buffer [1024];
error_t e;

  /* OUVERTURE DU FICHIER EN LECTURE-ECRITURE ET MODE APPEND */
  /* A priori le fichier n'existe pas et ce n'est pas un rpertoire. */
  if((e = open_file(name,
                    READING_AUTHORIZED,
                    WRITING_AUTHORIZED,
                    WRITING_AT_THE_END,
                    CREATE_FILE,
                    NOT_A_DIR,
                    &file_descr)) != 0)
  {
    fprintf(stderr, "Error while opening file %s : %s\n", name, string_of_error(e));
    return(EXIT_FAILURE);
  }

  do
  {
    /* LECTURE DE L'ENTREE STANDARD 
       les octets lus sont placs dans "buffer" */
    if((nb_bytes_read = read(STDIN_FILENO, buffer, sizeof(buffer))) == -1)
    {
      fprintf(stderr, "\"read (POSIX)\" : Error in reading file %s.\n", name);
      close_file(file_descr);
      return(EXIT_FAILURE);
    }
   
    /* ECRITURE DANS LE FICHIER */
    if((e = write_file(file_descr, buffer, nb_bytes_read)) != 0)
    {
      fprintf(stderr, "\"write_file\" : Error in writing. %s", string_of_error(e));
      close_file(file_descr);
      return(EXIT_FAILURE);
    }
  }
  while(nb_bytes_read != 0);
  
  close_file(file_descr);
  return(EXIT_SUCCESS);
}
/*-------------------------------*/
/*             mkdir_fs          */
/*-------------------------------*/
/*
Cette fonction permet de crer un rpertoire.
*/
int mkdir_fs(char * name)
{
fd file_descr;
error_t e;

  if((e = open_file(name, 
                    READING_AUTHORIZED, 
                    WRITING_AUTHORIZED, 
                    DOESNT_MATTER_HERE,
                    CREATE_FILE,
                    A_DIR,
                    &file_descr)) != 0)
  {
    fprintf(stderr, "\"mkdir_fs\" : Error in creating directory. %s\n", string_of_error(e));
    return(EXIT_FAILURE);
  }
  
  close_file(file_descr);
  return(EXIT_SUCCESS);
}
/*---------------------------------*/
/*             link_fs             */
/*---------------------------------*/
/*
Cette fonction permet de faire un lien sur un fichier.
*/
int link_fs(char * old_path, char * new_path)
{
error_t e;

  if((e = link_file(old_path, new_path)) != 0)
  {
    fprintf(stderr, "\"link_file\" Error in linking file %s. %s\n", old_path, string_of_error(e));
    return(EXIT_FAILURE);
  }

  return(EXIT_SUCCESS);
}
/*--------------------------------*/
/*            cp_fs               */
/*--------------------------------*/
/*
Cette fonction permet de copier un fichier dans un autre.
*/
int cp_fs(char * src, char * dest)
{
fd file_descr_src;
fd file_descr_dest;
char buffer [1024];
int nb_bytes_read;
error_t e;
  
  /* OUVERTURE DU FICHIER A COPIER EN LECTURE SEULE */
  if((e = open_file(src, 
                    READING_AUTHORIZED,
                    WRITING_PROHIBITED,
                    WRITING_AT_THE_BEGINNING, /* b_writing_at_the_end */
                    DOESNT_MATTER_HERE,       /* b_create*/
                    NOT_A_DIR,
                    &file_descr_src)) != 0)
  {
    fprintf(stderr, "\"open_file\" : Error in openning file %s. %s\n", src, string_of_error(e));
    return(EXIT_FAILURE);
  }
  
  /* CREATION/OUVERTURE DE LA COPIE EN ECRITURE SEULE */
  if((e = open_file(dest, 
                    READING_PROHIBITED,
                    WRITING_AUTHORIZED,
                    WRITING_AT_THE_BEGINNING, /* b_writing_at_the_end */
                    CREATE_FILE, /* b_create*/
                    NOT_A_DIR,
                    &file_descr_dest)) != 0)
  {
    fprintf(stderr, "\"open_file\" : Error in openning file %s. %s\n", dest, string_of_error(e));
    close_file(file_descr_src);
    return(EXIT_FAILURE);
  }
  
  do
  {
    /* LECTURE DANS L'ORIGINAL */
    if((e = read_file(file_descr_src, buffer, sizeof(buffer), &nb_bytes_read)) != 0)
    {
      fprintf(stderr, "\"read_file\" : Error in reading file %s. %s\n", src, string_of_error(e));
      close_file(file_descr_src);
      close_file(file_descr_dest);
      return(EXIT_FAILURE);
    }
    
    /* ECRITURE DES OCTETS LUS DANS LA COPIE */
    if((e = write_file(file_descr_dest, buffer, nb_bytes_read)) != 0)
    {
      fprintf(stderr, "\"write_file\" : Error in writing file %s. %s\n", dest, string_of_error(e));
      close_file(file_descr_src);
      close_file(file_descr_dest);
      return(EXIT_FAILURE);
    }
  }
  while(nb_bytes_read != 0);
    
  close_file(file_descr_src);
  close_file(file_descr_dest);
  return(EXIT_SUCCESS);  
}
/*----------------------------*/
/*            rm_fs           */
/*----------------------------*/
/*
Cette fonction permet de supprimer un lien sur un fichier.
*/
int rm_fs(char * path)
{
int dir_physical_block_number;
char * name;
error_t e;

  /* Comme ce que l'on veut, c'est connatre le rpertoire courant
     et le nom  supprimer dans celui-ci, on appelle la fonction
     "path_to_physical_block_number" avec l'argument 
     "FILE_NOT_ALREADY_CREATED". On rcupre dans "dir_physical_block_number"
     le numro du rpertoire courant et dans "name" le nom du lien 
      supprimer. 
  */
  if((e = path_to_physical_block_number(path, 
                                        &dir_physical_block_number,
                                        FILE_NOT_ALREADY_CREATED,
                                        &name)) != 0)
  {
    fprintf(stderr, "\"path_to_physical_block_number\" : Error. %s\n", string_of_error(e));
    return(EXIT_FAILURE);
  }

  if((e = remove_directory_entry(name, dir_physical_block_number)) != 0)
  {
    fprintf(stderr, "\"remove_directory_entry\" : Error in trying to remove %s from directory number %d. %s\n",
                     name, dir_physical_block_number, string_of_error(e));
    return(EXIT_FAILURE);
  }

  return(EXIT_SUCCESS);  
}
/*--------------------------------------------*/
/*                ls_fs                       */
/*--------------------------------------------*/
int ls_fs(char * dir)
{
fd file_descr;
fd file_descr_entry;
error_t e;
int b_end_of_dir;
dir_entry de;
struct fs_stat s_stat;
char type [4];        /* "dir" ou "reg"*/
char permissions [3]; /* "rw" ou "r-" ou "-w" ou "--" (rien) */

  /* OUVERTURE DU REPERTOIRE EN LECTURE SEULE */
  if((e = open_file(dir,
                    READING_AUTHORIZED,
                    WRITING_PROHIBITED,
                    WRITING_AT_THE_BEGINNING,
                    DOESNT_MATTER_HERE,
                    A_DIR,
                    &file_descr)) != 0)
  {
    fprintf(stderr, "\"open_file\" : error in trying to open %s : %s\n", dir, string_of_error(e));
    return(EXIT_FAILURE);
  }
  
  /* TANT QU'ON N'A PAS ATTEINT LA FIN DE L'ENTREE, ON LIT
     l'ENTREE COURANTE DANS LE REPERTOIRE */
  do
  {
    /* LECTURE D'UNE ENTREE DE REPERTOIRE */
    if((e = readdir_entry(file_descr, &de, &b_end_of_dir)) != 0)
    {
      fprintf(stderr, "\"readdir_entry\" : error in reading directory entry : %s\n", string_of_error(e));
      close_file(file_descr);
      return(EXIT_FAILURE);
    }

    /* FIN DU REPERTOIRE ? */
    if(b_end_of_dir == 1)
      break; 
    
    /* OUVERTURE DU FICHIER CORRESPONDANT A L'ENTREE LUE */
    if((e = open_file_from_physical_block_number(big_endian_to_u16(de.filenum),
                                                 READING_PROHIBITED,
                                                 WRITING_PROHIBITED,
                                                 WRITING_AT_THE_END,
                                                 &file_descr_entry)) != 0)
    {
      fprintf(stderr, "\"open_file_from_physical_block_number\" : %s\n", string_of_error(e));
      close_file(file_descr);
      return(EXIT_FAILURE);
    }

    /* RECUPERATION DES INFORMATIONS SUR L'ENTREE DE REPERTOIRE */
    if((e = stat_file(file_descr_entry, &s_stat)) != 0)
    {  
      fprintf(stderr, "\"stat_file\" : %s\n", string_of_error(e));
      close_file(file_descr);
      close_file(file_descr_entry);
      return(EXIT_FAILURE);
    }

    if(s_stat.st_type == TYPE_DIRECTORY)
      strcpy(type, "dir");

    else
      strcpy(type, "reg");

    if(s_stat.st_protect == 3)
      strcpy(permissions, "rw");

    else if(s_stat.st_protect == 2)
      strcpy(permissions, "r-");
    
    else if(s_stat.st_protect == 1)
      strcpy(permissions, "-w");
    
    else if(s_stat.st_protect == 0)
      strcpy(permissions, "--");
    
    /* AFFICHAGE SUR LA SORTIE STANDARD DES INFORMATIONS ET DU
       NOM DE L'ENTREE */
    printf("%d %s  %s  %d  %d  %s\n", file_descr_entry->nb_physical_block, 
                                      permissions, 
                                      type, 
                                      s_stat.st_size, 
                                      s_stat.st_links, 
                                      de.filename);
      
    close_file(file_descr_entry);
  }
  while(b_end_of_dir == 0);

  close_file(file_descr);

  return(EXIT_SUCCESS);
}
